import logging
import logging.config
import os
import sys
from datetime import datetime
from typing import Any, List, Literal, Optional

import click
import xarray as xr
import zarr

from copernicus_marine_client.catalogue_parser.catalogue_parser import (
    GEOCHUNKED_KEY,
    MOTU_KEY,
    OPENDAP_KEY,
    TIMECHUNKED_KEY,
    get_dataset_from_id,
    get_dataset_url_from_id,
    get_protocol_from_url,
    parse_catalogue,
    protocol_not_available_error,
)
from copernicus_marine_client.catalogue_parser.request_structure import (
    SubsetRequest,
    convert_motu_api_request_to_structure,
    subset_request_from_file,
)
from copernicus_marine_client.command_line_interface.exception_handler import (
    log_exception_and_exit,
)
from copernicus_marine_client.command_line_interface.group_login import (
    get_username_password,
)
from copernicus_marine_client.download_functions.download_motu import (
    download_motu,
)
from copernicus_marine_client.download_functions.download_opendap import (
    download_opendap,
)
from copernicus_marine_client.download_functions.download_zarr import (
    download_zarr,
    get_optimized_chunking,
)
from copernicus_marine_client.utils import (
    OVERWRITE_LONG_OPTION,
    OVERWRITE_OPTION_HELP_TEXT,
    OVERWRITE_SHORT_OPTION,
)

PROTOCOL_KEYS_ORDER = {
    "zarr": (TIMECHUNKED_KEY, GEOCHUNKED_KEY),
    "zarr-map": TIMECHUNKED_KEY,
    "zarr-timeserie": GEOCHUNKED_KEY,
    "opendap": OPENDAP_KEY,
    "motu": MOTU_KEY,
}


def to_command_line_interface_protocol(catalogue_protocol: str) -> str:
    return next(
        key
        for key, value in PROTOCOL_KEYS_ORDER.items()
        if value == catalogue_protocol
    )


def to_command_line_interface_protocols(
    catalogue_protocols: list[str],
) -> list[str]:
    return list(map(to_command_line_interface_protocol, catalogue_protocols))


CREDENTIALS_REQUIRED_PROTOCOLS = [
    "opendap",
    "motu",
]


@click.group()
def cli_group_subset() -> None:
    pass


@cli_group_subset.command(
    "subset",
    short_help="Downloads subsets of datasets as NetCDF files or Zarr stores",
    help="""
    Downloads subsets of datasets as NetCDF files or Zarr stores.
    Either one of 'dataset-id' or 'dataset-url' is required (can be found via the 'copernicus-marine describe' command).
    The arguments value passed individually through the CLI take precedence over the values from the "motu-api-request" option,
    which takes precedence over the ones from the "request-file" option

    Example:

    \b
    > copernicus-marine subset
    --dataset-id METOFFICE-GLO-SST-L4-NRT-OBS-SST-V2
    --variable analysed_sst --variable sea_ice_fraction
    --start-datetime 2021-01-01 --end-datetime 2021-01-02
    --minimal-longitude 0.0 --maximal-longitude 0.1
    --minimal-latitude 0.0 --maximal-latitude 0.1

    \b
    > copernicus-marine subset -i METOFFICE-GLO-SST-L4-NRT-OBS-SST-V2 -v analysed_sst -v sea_ice_fraction -t 2021-01-01 -T 2021-01-02 -x 0.0 -X 0.1 -y 0.0 -Y 0.1
    """,  # noqa
)
@click.option(
    "--dataset-url",
    "-u",
    type=str,
    help="The full dataset URL",
)
@click.option(
    "--dataset-id",
    "-i",
    type=str,
    help="The dataset id",
)
@click.option(
    "--username",
    type=str,
    envvar="COPERNICUS_MARINE_SERVICE_USERNAME",
    default=None,
    help="If not set, search for environment variable"
    + " COPERNICUS_MARINE_SERVICE_USERNAME"
    + ", or else look for configuration files, or else ask for user input",
)
@click.option(
    "--password",
    type=str,
    envvar="COPERNICUS_MARINE_SERVICE_PASSWORD",
    default=None,
    help="If not set, search for environment variable"
    + " COPERNICUS_MARINE_SERVICE_PASSWORD"
    + ", or else look for configuration files, or else ask for user input",
)
@click.option(
    "--variable",
    "-v",
    "variables",
    type=str,
    help="Specify dataset variables",
    multiple=True,
)
@click.option(
    "--minimal-longitude",
    "-x",
    type=float,
    help=(
        "Minimal longitude for the subset. "
        "The value will be reduced to the interval [-180; 360["
    ),
)
@click.option(
    "--maximal-longitude",
    "-X",
    type=float,
    help=(
        "Maximal longitude for the subset. "
        "The value will be reduced to the interval [-180; 360["
    ),
)
@click.option(
    "--minimal-latitude",
    "-y",
    type=click.FloatRange(min=-90, max=90),
    help="Minimal latitude for the subset. Requires a float within this range:",
)
@click.option(
    "--maximal-latitude",
    "-Y",
    type=click.FloatRange(min=-90, max=90),
    help="Maximal latitude for the subset. Requires a float within this range:",
)
@click.option(
    "--minimal-depth",
    "-z",
    type=click.FloatRange(min=0),
    help="Minimal depth for the subset. Requires a float within this range:",
)
@click.option(
    "--maximal-depth",
    "-Z",
    type=click.FloatRange(min=0),
    help="Maximal depth for the subset. Requires a float within this range:",
)
@click.option(
    "--start-datetime",
    "-t",
    type=click.DateTime(
        ["%Y", "%Y-%m-%d", "%Y-%m-%dT%H:%M:%S", "%Y-%m-%d %H:%M:%S"]
    ),
    help="The start datetime of the temporal subset. Caution: encapsulate date "
    + 'with " " to ensure valid expression for format "%Y-%m-%d %H:%M:%S"',
)
@click.option(
    "--end-datetime",
    "-T",
    type=click.DateTime(
        ["%Y", "%Y-%m-%d", "%Y-%m-%dT%H:%M:%S", "%Y-%m-%d %H:%M:%S"]
    ),
    help="The end datetime of the temporal subset. Caution: encapsulate date "
    + 'with " " to ensure valid expression for format "%Y-%m-%d %H:%M:%S"',
)
@click.option(
    "--output-directory",
    "-o",
    type=click.Path(),
    help="The destination folder for the downloaded files."
    + " Default is the current directory",
)
@click.option(
    "--config-file-directory",
    type=str,
    default=os.path.join(os.path.expanduser("~"), ".copernicus_marine_client"),
    help="Path to a directory where a configuration file is stored. Accepts "
    + ".copernicus_marine_client_credentials / .netrc or _netrc / "
    + ".motuclient-python.ini files",
)
@click.option(
    "--output-filename",
    "-f",
    type=click.Path(),
    help=(
        "Concatenate the downloaded data in the given file name "
        "(under the output directory). If "
        "the output-filename argument ends with '.nc' suffix, the file will be "
        "downloaded as a netCDF file."
    ),
)
@click.option(
    "--force-download",
    is_flag=True,
    default=False,
    help="Flag to skip confirmation before download",
)
@click.option(
    OVERWRITE_LONG_OPTION,
    OVERWRITE_SHORT_OPTION,
    is_flag=True,
    default=False,
    help=OVERWRITE_OPTION_HELP_TEXT,
)
@click.option(
    "--force-protocol",
    type=click.Choice(list(PROTOCOL_KEYS_ORDER.keys())),
    help="Force download through one of the available protocols",
)
@click.option(
    "--request-file",
    type=click.Path(),
    help="Option to pass a filename corresponding to a file containg CLI arguments. "
    "The file MUST follow the structure of dataclass 'SubsetRequest'. ",
)
@click.option(
    "--motu-api-request",
    type=str,
    help=(
        "Option to pass a complete MOTU api request as a string. "
        'Caution, user has to replace double quotes " with single '
        "quotes ' in the request"
    ),
)
@click.option(
    "--log-level",
    type=click.Choice(["DEBUG", "INFO", "WARN", "ERROR", "CRITICAL", "QUIET"]),
    default="INFO",
    help=(
        "Set the details printed to console by the command "
        "(based on standard logging library)."
    ),
)
@log_exception_and_exit
def subset(
    dataset_url: str,
    dataset_id: str,
    username: str,
    password: str,
    variables: Optional[List[str]],
    minimal_longitude: Optional[float],
    maximal_longitude: Optional[float],
    minimal_latitude: Optional[float],
    maximal_latitude: Optional[float],
    minimal_depth: Optional[float],
    maximal_depth: Optional[float],
    start_datetime: Optional[datetime],
    end_datetime: Optional[datetime],
    output_filename: Optional[str],
    force_protocol: Optional[str],
    request_file: Optional[str],
    output_directory: str,
    config_file_directory: str,
    motu_api_request: Optional[str],
    force_download: bool = False,
    overwrite_output_data: bool = False,
    log_level: str = "INFO",
):
    if log_level == "QUIET":
        logging.root.disabled = True
        logging.root.setLevel(level="CRITICAL")
    else:
        logging.root.setLevel(level=log_level)
    subset_request = SubsetRequest()
    if request_file:
        subset_request = subset_request_from_file(request_file)
    if motu_api_request:
        motu_api_subset_request = convert_motu_api_request_to_structure(
            motu_api_request
        )
        subset_request.update(motu_api_subset_request.__dict__)
    request_update_dict = {
        "dataset_url": dataset_url,
        "dataset_id": dataset_id,
        "variables": variables,
        "minimal_longitude": minimal_longitude,
        "maximal_longitude": maximal_longitude,
        "minimal_latitude": minimal_latitude,
        "maximal_latitude": maximal_latitude,
        "minimal_depth": minimal_depth,
        "maximal_depth": maximal_depth,
        "start_datetime": start_datetime,
        "end_datetime": end_datetime,
        "output_directory": output_directory,
        "output_filename": output_filename,
        "force_protocol": force_protocol,
    }
    subset_request.update(request_update_dict)
    if all(
        e is None
        for e in [
            subset_request.variables,
            subset_request.minimal_longitude,
            subset_request.maximal_longitude,
            subset_request.minimal_latitude,
            subset_request.maximal_latitude,
            subset_request.minimal_depth,
            subset_request.maximal_depth,
            subset_request.start_datetime,
            subset_request.end_datetime,
        ]
    ):
        logging.error(
            "The requested dataset is not subset, "
            "please use the native command instead."
        )
        sys.exit(1)
    # Specific treatment for default values:
    # In order to not overload arguments with default values
    if force_download:
        subset_request.force_download = force_download
    if overwrite_output_data:
        subset_request.overwrite = overwrite_output_data

    subset_function(
        username,
        password,
        subset_request,
        config_file_directory,
    )


def subset_function(
    username: str,
    password: str,
    subset_request: SubsetRequest,
    config_file_directory: str = os.path.join(
        os.path.expanduser("~"), ".copernicus_marine_client"
    ),
) -> str:
    def _flatten(x):
        if isinstance(x, (tuple, list)):
            return [a for i in x for a in _flatten(i)]
        else:
            return [x]

    catalogue = parse_catalogue()
    # ---Protocol section
    possible_protocols = (
        [p for p in list(PROTOCOL_KEYS_ORDER.values()) if isinstance(p, str)]
        if not subset_request.force_protocol
        else _flatten(PROTOCOL_KEYS_ORDER[subset_request.force_protocol])
    )
    if subset_request.force_protocol:
        logging.info(
            f"You forced selection of protocol: {subset_request.force_protocol}"
        )
    if not subset_request.dataset_url:
        if not subset_request.dataset_id:
            syntax_error = SyntaxError(
                "Must specify at least one of 'dataset_url' or 'dataset_id'"
            )
            logging.error(syntax_error)
            raise syntax_error
        protocol_keys_iterator = iter(possible_protocols)
        while not subset_request.dataset_url:
            try:
                protocol = next(protocol_keys_iterator)
            except StopIteration as exception:
                catalogue_available_protocols = get_dataset_from_id(
                    catalogue, subset_request.dataset_id
                ).get_available_protocols()
                available_protocols = to_command_line_interface_protocols(
                    catalogue_available_protocols
                )
                raise protocol_not_available_error(
                    subset_request.dataset_id, available_protocols
                ) from exception
            subset_request.dataset_url = get_dataset_url_from_id(
                catalogue, subset_request.dataset_id, protocol
            )
    else:
        protocol = get_protocol_from_url(subset_request.dataset_url)
    username, password = get_username_password(
        username,
        password,
        config_file_directory,
    )
    # --- Download redirection by protocol
    if subset_request.force_protocol and (
        protocol != PROTOCOL_KEYS_ORDER[subset_request.force_protocol]
        and protocol not in PROTOCOL_KEYS_ORDER[subset_request.force_protocol]
    ):
        attribute_error = AttributeError(
            f"Dataset url ({subset_request.dataset_url}) does not match forced "
            f"protocol ({PROTOCOL_KEYS_ORDER[subset_request.force_protocol]})!"
        )
        logging.error(attribute_error)
        raise attribute_error
    elif protocol in [TIMECHUNKED_KEY, GEOCHUNKED_KEY]:
        # Check if both timechunked and geochunked data are available
        url_timechunked, url_geochunked = (
            map(
                get_dataset_url_from_id,
                [catalogue] * 2,
                [subset_request.dataset_id] * 2,
                [TIMECHUNKED_KEY, GEOCHUNKED_KEY],
            )
            if subset_request.dataset_id
            else (None, None)
        )
        if (
            url_timechunked
            and url_geochunked
            and (subset_request.force_protocol in [None, "zarr"])
        ):
            subset_request.dataset_url = (
                url_timechunked
                if get_optimized_chunking(subset_request)
                else url_geochunked
            )
        logging.info("Download Zarr files through S3")
        output_name = download_zarr(
            username,
            password,
            subset_request,
        )

    elif protocol == OPENDAP_KEY:
        logging.info("Download through OPeNDAP")
        output_name = download_opendap(
            username,
            password,
            subset_request,
        )
    elif protocol == MOTU_KEY:
        logging.info("Download through MOTU")
        output_name = download_motu(
            username,
            password,
            subset_request,
            catalogue=catalogue,
        )
    elif not protocol:
        key_error = KeyError(
            f"The requested dataset '{subset_request.dataset_id}' does not have "
            f"{possible_protocols} url available"
        )
        logging.error(key_error)
        raise
    else:
        key_error = KeyError(
            f"Protocol {protocol} not handled by subset command"
        )
        logging.error(key_error)
        raise key_error
    return output_name


def open_dataset(
    filepath: Optional[str] = None,
    engine: Literal["netcdf4", "h5netcdf", "pydap", "zarr", None] = None,
    username: Optional[str] = None,
    password: Optional[str] = None,
    subset_request: Optional[SubsetRequest] = None,
    out_type: Literal["xarray", "zarr", "pandas"] = "xarray",
) -> Any:
    """
    Function to open a dataset as a python object.
    Input:
    Either the path to a dataset or a SubsetRequest with the information
    necessary to download a dataset.
    The user may need to specify the "engine" needed by xarray to open the dataset.
    Output:
    Return the dataset as either:
        - a xarray dataset
        - a zarr ZarrStore object
        - a pandas dataframe
    """

    def _open_dataset_as_xarray(
        filepath: Optional[str],
        username: Optional[str],
        password: Optional[str],
        subset_request: Optional[SubsetRequest],
        engine: Literal["netcdf4", "h5netcdf", "pydap", "zarr", None],
    ) -> xr.Dataset:
        if filepath:
            if filepath.split(".")[-1] == "nc" and engine is None:
                engine = "netcdf4"
            elif filepath.split(".")[-1] == "zarr" and engine is None:
                engine = "zarr"
            elif engine is None:
                name_error = NameError(
                    f"Dataset name ({filepath}) has non-explicit or unmapped extension "
                    "please specify an engine to open it."
                )
                logging.error(name_error)
                raise name_error
            store_name = filepath.rsplit(".", maxsplit=1)[0] + ".zarr"
            dataset = xr.open_dataset(filepath, engine=engine)
        elif subset_request and username and password:
            output_name = subset_function(username, password, subset_request)
            if output_name.split(".")[-1] == "nc" and engine is None:
                engine = "netcdf4"
            elif output_name.split(".")[-1] == "zarr" and engine is None:
                engine = "zarr"
            else:
                name_error = NameError(
                    f"Output name ({output_name}) has non-explicit or unmapped "
                    "extension please specify an engine to open it."
                )
                logging.error(name_error)
                raise name_error
            store_name = output_name.rsplit(".", maxsplit=1)[0] + ".zarr"
            dataset = xr.open_dataset(output_name, engine=engine)
        else:
            attribute_error = AttributeError(
                "Not enough inputs, specify either a filepath "
                "or pass a valid subset_request with credentials"
            )
            raise attribute_error
        return dataset, store_name

    def _open_dataset_as_out_type(
        xarray_dataset: xr.Dataset, out_type: str, store_name: str
    ):
        if out_type == "zarr":
            store = zarr.DirectoryStore(store_name)
            xarray_dataset.to_zarr(store, mode="w")
            return zarr.open(store)
        elif out_type == "xarray":
            return xarray_dataset
        elif out_type == "pandas":
            return xarray_dataset.to_dataframe()

    xarray_dataset, store_name = _open_dataset_as_xarray(
        filepath, username, password, subset_request, engine
    )
    return _open_dataset_as_out_type(xarray_dataset, out_type, store_name)



import os
import re
from multiprocessing.pool import ThreadPool
from pathlib import Path
from typing import List, Optional, Tuple

import boto3
import botocore
from numpy import append, arange
from tqdm import tqdm

from copernicus_marine_client.catalogue_parser.request_structure import (
    NativeRequest,
)
from copernicus_marine_client.download_functions.download_native import (
    download_native,
)
from copernicus_marine_client.utils import get_unique_filename

MARINE_DATA_LAKE_LISTING_ENDPOINT = "https://marine.copernicus.eu"
MARINE_DATA_LAKE_LISTING_NATIVE_BUCKET = "mdl-native-list"
MARINE_DATA_LAKE_DOWNLOAD_NATIVE_BUCKET = "mdl-native"
MARINE_DATA_LAKE_S3_LISTING_NATIVE_ROOT_URI = (
    f"{MARINE_DATA_LAKE_LISTING_NATIVE_BUCKET}/native"
)
MARINE_DATA_LAKE_S3_DOWNLOAD_NATIVE_ROOT_URI = (
    f"{MARINE_DATA_LAKE_DOWNLOAD_NATIVE_BUCKET}/native"
)


def download_s3native(
    username: str,
    password: str,
    native_request: NativeRequest,
) -> str:
    filenames_in, filenames_out, endpoint_url = download_native(
        username,
        password,
        native_request,
        download_header,
        create_filenames_out,
    )
    return download_files(endpoint_url, filenames_in, filenames_out)


def download_files(
    endpoint_url: str,
    filenames_in: List[str],
    filenames_out: List[str],
) -> str:
    pool = ThreadPool()
    nfiles_per_process, nfiles = 1, len(filenames_in)
    indexes = append(
        arange(0, nfiles, nfiles_per_process, dtype=int),
        nfiles,
    )
    groups_in_files = [
        filenames_in[indexes[i] : indexes[i + 1]]
        for i in range(len(indexes) - 1)
    ]
    groups_out_files = [
        filenames_out[indexes[i] : indexes[i + 1]]
        for i in range(len(indexes) - 1)
    ]

    for groups_out_file in groups_out_files:
        parent_dir = Path(groups_out_file[0]).parent
        if not os.path.isdir(parent_dir):
            os.makedirs(parent_dir)

    download_summary_list = pool.imap(
        _download_files,
        zip(
            [endpoint_url] * len(groups_in_files),
            groups_in_files,
            groups_out_files,
        ),
    )
    list(tqdm(download_summary_list, total=len(groups_in_files)))
    download_summary = "Download complete"
    return download_summary


def download_header(
    data_path: str, regex: Optional[str], username: str, password: str
) -> Tuple[str, str, list[str], float]:
    (endpoint_url, path) = parse_s3native_dataset_url(data_path)
    message = "You requested the download of the following files:\n"
    filenames, sizes, total_size = [], [], 0.0
    raw_filenames = list_files_on_marine_data_lake_s3(path)
    filename_filtered = []
    for filename, size in raw_filenames:
        if not regex or re.match(regex, filename):
            filenames += [filename]
            sizes += [float(size)]
            total_size += float(size)
            filename_filtered.append((filename, size))

    for filename, size in filename_filtered[:20]:
        message += str(filename)
        message += f" - {format_file_size(float(size))}\n"
    if len(filenames) > 20:
        message += f"Printed 20 out of {len(filenames)} files\n"
    message += (
        f"\nTotal size of the download: {format_file_size(total_size)}\n\n"
    )
    return (message, endpoint_url, filenames, total_size)


def list_files_on_marine_data_lake_s3(
    path: str,
    files_already_found: list[tuple[str, int]] = [],
) -> list[tuple[str, int]]:
    # In order to list S3 native files, we are using CDN functionnalities using the following endpoint_url and path  # noqa
    path = path.replace(
        f"s3://{MARINE_DATA_LAKE_S3_DOWNLOAD_NATIVE_ROOT_URI}",
        f"s3://{MARINE_DATA_LAKE_S3_LISTING_NATIVE_ROOT_URI}",
    )

    if not path.endswith("/"):
        path += "/"

    s3_session = boto3.session.Session()
    s3_client = s3_session.client(
        "s3",
        config=botocore.config.Config(
            # Configures to use subdomain/virtual calling format.
            s3={"addressing_style": "virtual"},
            signature_version=botocore.UNSIGNED,
        ),
        endpoint_url=MARINE_DATA_LAKE_LISTING_ENDPOINT,
    )

    prefix = path.replace(
        f"s3://{MARINE_DATA_LAKE_LISTING_NATIVE_BUCKET}/", ""
    )

    response = s3_client.list_objects_v2(
        Bucket=MARINE_DATA_LAKE_LISTING_NATIVE_BUCKET, Prefix=prefix
    )
    s3_objects = response["Contents"]

    for s3_object in s3_objects:
        files_already_found.extend(
            [
                (
                    f"s3://{MARINE_DATA_LAKE_DOWNLOAD_NATIVE_BUCKET}/"
                    + s3_object["Key"],
                    s3_object["Size"],
                )
            ]
        )
    return files_already_found


def _download_files(
    tuple_s3native_filename: Tuple[str, list[str], list[str]],
) -> str:
    def _s3native_file_download(
        endpoint_url: str, file_in: str, file_out: str
    ) -> str:
        """
        Download ONE file and return a string of the result
        """
        s3_session = boto3.session.Session()
        s3_client = s3_session.client(
            "s3",
            config=botocore.config.Config(
                # Configures to use subdomain/virtual calling format.
                s3={"addressing_style": "virtual"},
                signature_version=botocore.UNSIGNED,
            ),
            endpoint_url=endpoint_url,
        )

        s3_client.download_file(
            MARINE_DATA_LAKE_DOWNLOAD_NATIVE_BUCKET,
            file_in.replace("s3://mdl-native/", ""),
            file_out,
        )

        return f"File {file_out} created\n"

    endpoint_url, filenames_in, filenames_out = tuple_s3native_filename
    download_summary = ""
    for file_in, file_out in zip(filenames_in, filenames_out):
        download_summary += _s3native_file_download(
            endpoint_url, file_in, file_out
        )
    return download_summary


# /////////////////////////////
# --- Tools
# /////////////////////////////


def parse_s3native_dataset_url(data_path: str) -> Tuple[str, str]:
    endpoint_url, path = data_path.split("/mdl-native/", maxsplit=1)
    path = "s3://mdl-native/" + path
    return (endpoint_url, path)


def create_filenames_out(
    filenames_in: list[str],
    overwrite: bool,
    output_directory: str = "",
    no_directories=False,
) -> list[str]:
    filenames_out = []
    for filename_in in filenames_in:
        filename_out = f"{output_directory}/"
        if no_directories:
            filename_out += filename_in.split("/")[-1]
        elif filename_in.startswith(
            f"s3://{MARINE_DATA_LAKE_S3_DOWNLOAD_NATIVE_ROOT_URI}{os.sep}"
        ):
            filename_out += filename_in.split(
                f"s3://{MARINE_DATA_LAKE_S3_DOWNLOAD_NATIVE_ROOT_URI}{os.sep}"
            )[-1]
        elif filename_in.startswith(
            f"s3://{MARINE_DATA_LAKE_S3_LISTING_NATIVE_ROOT_URI}{os.sep}"
        ):
            filename_out += filename_in.split(
                f"s3://{MARINE_DATA_LAKE_S3_LISTING_NATIVE_ROOT_URI}{os.sep}"
            )[-1]

        filename_out = get_unique_filename(
            filepath=filename_out, overwrite_option=overwrite
        )

        filenames_out.append(filename_out)
    return filenames_out


def format_file_size(
    size: float, decimals: int = 2, binary_system: bool = False
) -> str:
    if binary_system:
        units: list[str] = [
            "B",
            "KiB",
            "MiB",
            "GiB",
            "TiB",
            "PiB",
            "EiB",
            "ZiB",
        ]
        largest_unit: str = "YiB"
        step: int = 1024
    else:
        units = ["B", "kB", "MB", "GB", "TB", "PB", "EB", "ZB"]
        largest_unit = "YB"
        step = 1000

    for unit in units:
        if size < step:
            return ("%." + str(decimals) + "f %s") % (size, unit)
        size /= step

    return ("%." + str(decimals) + "f %s") % (size, largest_unit)



import json
import logging
import logging.config
import os

import click

from copernicus_marine_client.command_line_interface.group_describe import (
    cli_group_describe,
)
from copernicus_marine_client.command_line_interface.group_login import (
    cli_group_login,
)
from copernicus_marine_client.command_line_interface.group_native import (
    cli_group_native,
)
from copernicus_marine_client.command_line_interface.group_subset import (
    cli_group_subset,
)

log_config_dict = json.load(
    open(os.path.join(os.path.dirname(__file__), "..", "logging_conf.json"))
)
logging.config.dictConfig(log_config_dict)


@click.command(
    cls=click.CommandCollection,
    sources=[
        cli_group_describe,
        cli_group_login,
        cli_group_subset,
        cli_group_native,
    ],
)
@click.version_option(
    None, "-V", "--version", package_name="copernicus-marine-client"
)
def command_line_interface():
    pass


if __name__ == "__main__":
    command_line_interface()



"""
.
"""

import importlib.metadata

__version__ = importlib.metadata.version("copernicus-marine-client")

from copernicus_marine_client.catalogue_parser.catalogue_parser import (
    CopernicusMarineCatalogue as CopernicusMarineCatalogue,
)
from copernicus_marine_client.catalogue_parser.catalogue_parser import (
    CopernicusMarineDatasetCoordinates as CopernicusMarineDatasetCoordinates,
)
from copernicus_marine_client.catalogue_parser.catalogue_parser import (
    CopernicusMarineDatasetService as CopernicusMarineDatasetService,
)
from copernicus_marine_client.catalogue_parser.catalogue_parser import (
    CopernicusMarineDatasetVariable as CopernicusMarineDatasetVariable,
)
from copernicus_marine_client.catalogue_parser.catalogue_parser import (
    CopernicusMarineProduct as CopernicusMarineProduct,
)
from copernicus_marine_client.catalogue_parser.catalogue_parser import (
    CopernicusMarineProductDataset as CopernicusMarineProductDataset,
)
from copernicus_marine_client.catalogue_parser.catalogue_parser import (
    CopernicusMarineProductProvider as CopernicusMarineProductProvide,
)
from copernicus_marine_client.catalogue_parser.catalogue_parser import (
    parse_catalogue as fetch_catalogue,
    get_all_dataset_ids as get_all_dataset_ids,
    merge_catalogues as merge_catalogues,
    parse_dissemination_unit_catalogue as fetch_dissemination_unit_catalogue,
    parse_marine_data_store_catalogue as fetch_marine_data_store_catalogue,
)


from copernicus_marine_client.catalogue_parser.request_structure import (
    NativeRequest as NativeRequest,
    native_request_from_file as native_request_from_file
    )
from copernicus_marine_client.catalogue_parser.request_structure import (
    SubsetRequest as SubsetRequest,
    subset_request_from_file as subset_request_from_file, 
    convert_motu_api_request_to_structure as convert_motu_api_request_to_structure
    )
from copernicus_marine_client.command_line_interface.group_native import (
    PROTOCOL_KEYS_ORDER as NATIVE_PROTOCOL_KEYS_ORDER,
)
from copernicus_marine_client.command_line_interface.group_native import (
    native_function as download_native,
)
from copernicus_marine_client.command_line_interface.group_subset import (
    PROTOCOL_KEYS_ORDER as SUBSET_PROTOCOL_KEYS_ORDER,
)
from copernicus_marine_client.command_line_interface.group_subset import open_dataset as open_dataset
from copernicus_marine_client.command_line_interface.group_subset import (
    subset_function as download_subset,
)
from copernicus_marine_client.download_functions.download_ftp import download_ftp as download_ftp
from copernicus_marine_client.download_functions.download_motu import download_motu as download_motu
from copernicus_marine_client.download_functions.download_opendap import (
    download_opendap as download_opendap,
)
from copernicus_marine_client.download_functions.download_s3native import (
    download_s3native as download_s3native,
)
from copernicus_marine_client.download_functions.download_zarr import download_zarr as download_zarr
from copernicus_marine_client.download_functions.download_zarr import (
    get_optimized_chunking as get_optimized_chunking,
)



import logging
import logging.config
import os
from typing import Literal, Optional, Tuple

import click

from copernicus_marine_client.configuration_files_creator import (
    check_copernicus_marine_credentials,
)
from copernicus_marine_client.configuration_files_creator import (
    main as configuration_files_creator,
)
from copernicus_marine_client.configuration_files_creator import (
    retrieve_credential_from_config_files,
)


@click.group()
def cli_group_login() -> None:
    pass


@cli_group_login.command(
    "login",
    help="""
    This command check the copernicus-marine credentials provided by the user
    and creates a configuration file with the encoded credentials if the check is valid.
    It then stores the configuration file in a directory that can be specified by
    the user.
    If the user specified a different 'config_file_directory' from default one
    ($HOME/.copernicus_marine_client), it needs to be passed also to the download
    commands.

    Examples:

    Case 1 (Recommended):

    With environment variables COPERNICUS_MARINE_SERVICE_USERNAME &
    COPERNICUS_MARINE_SERVICE_PASSWORD specified:

    > copernicus-marine login

    Case 2:

    > copernicus-marine login \n
    < Username: [USER-INPUT] \n
    < Password: [USER-INPUT]

    Case 3:

    > copernicus-marine login --username JOHN_DOE --password SECRETPASSWORD

    Case 4: Specific directory for config_files

    > copernicus-marine login --config-file-directory USER/SPECIFIED/PATH
        """,
)
@click.option(
    "--username",
    prompt="username",
    envvar="COPERNICUS_MARINE_SERVICE_USERNAME",
    hide_input=False,
    help="If not set, search for environment variable"
    + " COPERNICUS_MARINE_SERVICE_USERNAME"
    + ", or else ask for user input",
)
@click.option(
    "--password",
    prompt="password",
    envvar="COPERNICUS_MARINE_SERVICE_PASSWORD",
    hide_input=True,
    help="If not set, search for environment variable"
    + " COPERNICUS_MARINE_SERVICE_PASSWORD"
    + ", or else ask for user input",
)
@click.option(
    "--config-file-directory",
    type=str,
    default=os.path.join(os.path.expanduser("~"), ".copernicus_marine_client"),
    help="Path to the directory where the configuration file is stored",
)
@click.option(
    "--overwrite-configuration-file",
    "-overwrite",
    is_flag=True,
    default=False,
    help="Flag to skip confirmation before overwriting configuration file",
)
@click.option(
    "--verbose",
    type=click.Choice(["DEBUG", "INFO", "WARN", "ERROR", "CRITICAL", "QUIET"]),
    default="INFO",
    help=(
        "Set the details printed to console by the command "
        "(based on standard logging library)."
    ),
)
def login(
    username: str,
    password: str,
    config_file_directory: str,
    overwrite_configuration_file: bool,
    verbose: str = "INFO",
) -> None:
    if verbose == "QUIET":
        logging.root.disabled = True
        logging.root.setLevel(level="CRITICAL")
    else:
        logging.root.setLevel(level=verbose)
    check_copernicus_marine_credentials(username, password)
    configuration_files_creator(
        username=username,
        password=password,
        config_file_directory=config_file_directory,
        overwrite_configuration_file=overwrite_configuration_file,
    )
    logging.info(f"Configuration files stored in {config_file_directory}")


def get_credential(
    credential: Optional[str],
    credential_type: Literal["username", "password"],
    hide_input: bool,
    config_file_directory: str,
) -> str:
    if not credential:
        credential = retrieve_credential_from_config_files(
            credential_type=credential_type,
            config_file_directory=config_file_directory,
            host="my.cmems_du.eu",  # Same credentials for all hosts
        )
        if not credential:
            credential = click.prompt(credential_type, hide_input=hide_input)
            if not credential:
                raise ValueError(f"{credential} cannot be None")
    else:
        logging.debug(
            "Credentials loaded from function arguments or environment variable"
        )
    return credential


def get_username_password(
    username: Optional[str],
    password: Optional[str],
    config_file_directory: str,
) -> Tuple[str, str]:
    username = get_credential(
        username,
        "username",
        hide_input=False,
        config_file_directory=config_file_directory,
    )
    password = get_credential(
        password,
        "password",
        hide_input=True,
        config_file_directory=config_file_directory,
    )
    result_check = check_copernicus_marine_credentials(username, password)
    if result_check.error:
        logging.warning(
            "Invalid credentials, your download will not be authenticated"
        )
    return (username, password)


if __name__ == "__main__":
    cli_group_login()
